#include <stdio.h>
#include <FL/Fl.H>
#include <FL/x.H>
#include <FL/Fl_Gl_Window.H>
#ifdef __APPLE__
#  include <OpenGL/gl3.h>
#  include <OpenGL/OpenGL.h>
#else
#  include <GL/glew.h>
#  include <dlfcn.h>
#endif

static void glew_initialize() {
#ifndef __APPLE__
  static bool first = true;
  if (first) {
    first = false;   
    int major = 1;
    sscanf((const char *)glGetString(GL_VERSION), "%d", &major);
    if (major >= 3) { // we are using OpenGL V3 or more
      typedef GLenum (*pftype)(void);
      pftype pf = (pftype)dlsym(NULL, "glewInit");
      GLenum err = 1;
      if (pf) err = pf(); // calling glewInit() is compulsory to use OpenGL V3 API
      printf("glewInit=%p err=%d major=%d\n",pf,err,major);
    }
  }
#endif
}

class SimpleGL3Window : public Fl_Gl_Window {
  GLuint shaderProgram;
  GLuint vertexArrayObject;
  GLuint vertexBuffer;
  
  GLint positionUniform;
  GLint colourAttribute;
  GLint positionAttribute;
public:
  SimpleGL3Window(int w, int h) :  Fl_Gl_Window(w, h) {
#ifdef __APPLE__
    static int array[] = { // equivalent to FL_RGB8 | FL_DOUBLE
      kCGLPFADoubleBuffer,
      kCGLPFAOpenGLProfile,
      kCGLOGLPVersion_3_2_Core,
      //kCGLOGLPVersion_GL4_Core,
      0
    };
    mode(array);
#else
    mode(FL_RGB8 | FL_DOUBLE);
#endif
    shaderProgram = 0;
  }
  void draw() {
    if (!shaderProgram) {
      glew_initialize();
      GLuint  vs;
      GLuint  fs;
      int Mslv, mslv; // major and minor version numbers of the shading language
      sscanf((char*)glGetString(GL_SHADING_LANGUAGE_VERSION), "%d.%d", &Mslv, &mslv);
      //fprintf(stderr,"slv=%d.%d\n",Mslv, mslv);
      const char *vss_format="#version %d%d\n\
      uniform vec2 p;\
      in vec4 position;\
      in vec4 colour;\
      out vec4 colourV;\
      void main (void)\
      {\
      colourV = colour;\
      gl_Position = vec4(p, 0.0, 0.0) + position;\
      }";
      char vss_string[300]; const char *vss = vss_string;
      sprintf(vss_string, vss_format, Mslv, mslv);
      const char *fss_format="#version %d%d\n\
      in vec4 colourV;\
      out vec4 fragColour;\
      void main(void)\
      {\
      fragColour = colourV;\
      }";
      char fss_string[200]; const char *fss = fss_string;
      sprintf(fss_string, fss_format, Mslv, mslv);
      GLint err; GLchar LOG[1000]; GLsizei length;
      vs = glCreateShader(GL_VERTEX_SHADER);
      glShaderSource(vs, 1, &vss, NULL);
      glCompileShader(vs);
      glGetShaderiv(vs, GL_COMPILE_STATUS, &err);
      if (err != GL_TRUE) {
	glGetShaderInfoLog(vs, sizeof(LOG), &length, LOG); 
	fprintf(stderr,"vs ShaderInfoLog=%s\n",LOG);
	}
      fs = glCreateShader(GL_FRAGMENT_SHADER);
      glShaderSource(fs, 1, &fss, NULL);
      glCompileShader(fs);
      glGetShaderiv(fs, GL_COMPILE_STATUS, &err);
      if (err != GL_TRUE) {
	glGetShaderInfoLog(fs, sizeof(LOG), &length, LOG); 
	fprintf(stderr,"fs ShaderInfoLog=%s\n",LOG);
	}     
      // Attach the shaders
      shaderProgram = glCreateProgram();
      glAttachShader(shaderProgram, vs);
      glAttachShader(shaderProgram, fs);
      glBindFragDataLocation(shaderProgram, 0, "fragColour");
      glLinkProgram(shaderProgram);
      glGetProgramiv(shaderProgram, GL_LINK_STATUS, &err);
      if (err != GL_TRUE) {
        glGetProgramInfoLog(shaderProgram, sizeof(LOG), &length, LOG);
        fprintf(stderr,"link log=%s\n", LOG);
      }
      // Get pointers to uniforms and attributes
      positionUniform = glGetUniformLocation(shaderProgram, "p");
      colourAttribute = glGetAttribLocation(shaderProgram, "colour");
      positionAttribute = glGetAttribLocation(shaderProgram, "position");
      glDeleteShader(vs);
      glDeleteShader(fs);
      //printf("positionUniform: %i, colourAttribute: %i, positionAttribute: %i\n",positionUniform,colourAttribute,positionAttribute);
      
      // Upload vertices (1st four values in a row) and colours (following four values)
      GLfloat vertexData[]= { -0.5,-0.5,0.0,1.0,   1.0,0.0,0.0,1.0,
        -0.5, 0.5,0.0,1.0,   0.0,1.0,0.0,1.0,
        0.5, 0.5,0.0,1.0,   0.0,0.0,1.0,1.0,
        0.5,-0.5,0.0,1.0,   1.0,1.0,1.0,1.0};
      glGenVertexArrays(1, &vertexArrayObject);
      glBindVertexArray(vertexArrayObject);
      
      glGenBuffers(1, &vertexBuffer);
      glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
      glBufferData(GL_ARRAY_BUFFER, 4*8*sizeof(GLfloat), vertexData, GL_STATIC_DRAW);
      
      glEnableVertexAttribArray((GLuint)positionAttribute);
      glEnableVertexAttribArray((GLuint)colourAttribute  );
      glVertexAttribPointer((GLuint)positionAttribute, 4, GL_FLOAT, GL_FALSE, 8*sizeof(GLfloat), 0);
      glVertexAttribPointer((GLuint)colourAttribute  , 4, GL_FLOAT, GL_FALSE, 8*sizeof(GLfloat), (char*)0+4*sizeof(GLfloat));
    }
    else if ((!valid())) {
#if FL_MAJOR_VERSION*1000+FL_MINOR_VERSION*100+FL_PATCH_VERSION > 10304
      glViewport(0, 0, pixel_w(), pixel_h());
#else
      glViewport(0, 0, w(), h());
#endif
    }
    glClearColor(0.08, 0.8, 0.8, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);
    glUseProgram(shaderProgram);
    GLfloat p[]={0,0};
    glUniform2fv(positionUniform, 1, (const GLfloat *)&p);
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
  }
  int handle(int event) {
    if (event == FL_PUSH) {
      static float factor = 1.1;
      GLfloat data[4];
      glGetBufferSubData(GL_ARRAY_BUFFER, 0, 4*sizeof(GLfloat), data);
      if (data[0] < -0.88 || data[0] > -0.5) factor = 1/factor;
      data[0] *= factor;
      glBufferSubData(GL_ARRAY_BUFFER, 0, 4*sizeof(GLfloat), data);
      glGetBufferSubData(GL_ARRAY_BUFFER, 24*sizeof(GLfloat), 4*sizeof(GLfloat), data);
      data[0] *= factor;
      glBufferSubData(GL_ARRAY_BUFFER, 24*sizeof(GLfloat), 4*sizeof(GLfloat), data);
      redraw();
      return 1;
    }
    return Fl_Gl_Window::handle(event);
  }
};


int main(int argc, char **argv)
{
#if defined(__APPLE__) && FL_MAJOR_VERSION*1000+FL_MINOR_VERSION*100+FL_PATCH_VERSION > 10304
  Fl::use_high_res_GL(1);
#endif
  SimpleGL3Window *win = new SimpleGL3Window(500, 500);
  win->end();
  win->resizable(win);
  win->label("Click to reshape");
  win->show();
  Fl::run();
}
